from dateutil import parser
from typing import Optional, List, TYPE_CHECKING

from ..util import prettify, get_optional, get_optional_list
from ..enums import ChatChannelType, ChatMessageType
from .current_user_attributes import ChatChannelUserAttributes
from .user import UserCompact


if TYPE_CHECKING:
    from datetime import datetime


class ChatChannel:
    """
    Represents an individual chat "channel" in the game.

    **Attributes**

    channel_id: :class:`int`

    name: :class:`str`

    description: Optional[:class:`str`]

    icon: Optional[:class:`str`]
        display icon for the channel

    type: :class:`ChatChannelType`
        Below are the channel types and their permission checks for joining/messaging:
        PUBLIC
            None
        PRIVATE
            is player in the allowed groups? (channel.allowed_groups)
        MULTIPLAYER
            is player currently in the mp game?
        SPECTATOR
            None
        TEMPORARY
            deprecated
        PM
            For PMs, two factors are taken into account:
            Is either user blocking the other? If so, deny.
            Does the target only accept PMs from friends? Is the current user a friend? If not, deny.
        GROUP
            is player in channel? (user_channels)
        ANNOUNCE
            is user in the announce group?

    moderated: :class:`bool`
        user can't send message when the value is `True`

    uuid: Optional[:class:`str`]
        value from requests that is relayed back to the sender.

    current_user_attributes: Optional[:class:`ChatChannelUserAttributes`]
        only present on some responses

    last_message_id: Optional[:class:`int`]
        message_id of last known message (only returned in presence responses)

    recent_messages: Optional[List[:class:`ChatMessage`]]
        [DEPRECATED] up to 50 most recent messages

    users: Optional[List[:class:`int`]]
        list of user ids that are in the channel (not included for PUBLIC channels).
    """

    __slots__ = (
        "channel_id",
        "name",
        "description",
        "icon",
        "type",
        "moderated",
        "uuid",
        "current_user_attributes",
        "last_message_id",
        "recent_messages",
        "users",
    )

    def __init__(self, data):
        self.channel_id: int = data["channel_id"]
        self.name: str = data["name"]
        self.description: Optional[str] = data.get("description")
        self.icon: Optional[str] = data.get("icon")
        self.type: ChatChannelType = ChatChannelType(data["type"])
        self.moderated: bool = data["moderated"]

        self.uuid: Optional[str] = data.get("uuid")
        self.current_user_attributes: Optional[ChatChannelUserAttributes] = get_optional(
            data, "current_user_attributes", ChatChannelUserAttributes
        )
        self.last_message_id: Optional[int] = data.get("last_message_id")
        self.recent_messages: Optional[List[ChatMessage]] = get_optional_list(data, "recent_messages", ChatMessage)
        self.users: Optional[List[int]] = data.get("users")

    def __repr__(self):
        return prettify(self, "name")


class ChatMessage:
    """
    Represents an individual Message within a :class:`ChatChannel`.

    **Attributes**

    channel_id: :class:`int`
        channel_id of where the message was sent

    content: :class:`str`
        message content

    is_action: :class:`bool`
        was this an action? i.e. /me dances

    message_id: :class:`int`
        unique identifier for message

    sender_id: :class:`int`
        user_id of the sender

    timestamp: :class:`datetime.datetime`
        when the message was sent

    type: :class:`ChatMessageType`

    uuid: Optional[:class:`str`]
        message identifier originally sent by client

    sender: Optional[:class:`UserCompact`]
        embedded :class:`UserCompact` object to save additional api lookups
    """

    __slots__ = (
        "channel_id",
        "content",
        "is_action",
        "message_id",
        "sender_id",
        "timestamp",
        "type",
        "uuid",
        "sender",
    )

    def __init__(self, data):
        self.channel_id: int = data["channel_id"]
        self.content: str = data["content"]
        self.is_action: bool = data["is_action"]
        self.message_id: int = data["message_id"]
        self.sender_id: int = data["sender_id"]
        self.timestamp: datetime = parser.parse(data["timestamp"])
        self.type: ChatMessageType = ChatMessageType(data["type"])
        self.uuid: Optional[str] = data.get("uuid")
        self.sender: Optional[UserCompact] = get_optional(data, "sender", UserCompact)

    def __repr__(self):
        return prettify(self, "sender_id" if self.sender is None else "sender", "content")



from typing import Optional

from ..util import prettify, get_optional
from ..enums import GameModeStr


class Achievement:
    """
    **Attributes**

    icon_url: :class:`str`

    id: :class:`int`

    name: :class:`str`

    grouping: :class:`str`

    ordering: :class:`int`

    slug: :class:`str`

    description: :class:`str`

    mode: Optional[:class:`GameModeStr`]

    instructions: Optional[:class:`str`]
    """

    __slots__ = (
        "icon_url",
        "id",
        "name",
        "grouping",
        "ordering",
        "slug",
        "description",
        "mode",
        "instructions",
    )

    def __init__(self, data):
        self.icon_url: str = data["icon_url"]
        self.id: int = data["id"]
        self.name: str = data["name"]
        self.grouping: str = data["grouping"]
        self.ordering: int = data["ordering"]
        self.slug: str = data["slug"]
        self.description: str = data["description"]
        self.mode: GameModeStr = get_optional(data, "mode", GameModeStr)
        self.instructions: Optional[str] = data["instructions"]

    def __repr__(self):
        return prettify(self, "name", "description")



from dateutil import parser
from typing import Dict, Optional, List, TYPE_CHECKING

from .beatmap import Beatmapset
from .user import UserStatistics
from ..util import prettify, get_optional, get_optional_list


if TYPE_CHECKING:
    from datetime import datetime


class Rankings:
    """
    **Attributes**

    beatmapsets: Optional[List[:class:`Beatmapset`]]
        The list of beatmaps in the requested spotlight for the given mode;
        only available if type is charts

    cursor: Dict
        To be used to query the next page

    ranking: List[:class:`UserStatistics`]
        Score details ordered by rank in descending order.

    spotlight: Optional[:class:`Spotlight`]
        Spotlight details; only available if type is charts

    total: :class:`int`
        An approximate count of ranks available
    """

    __slots__ = ("beatmapsets", "cursor", "ranking", "spotlight", "total")

    def __init__(self, data):
        self.cursor: Dict = data["cursor"]
        self.ranking: List[UserStatistics] = list(map(UserStatistics, data["ranking"]))
        self.total: int = data["total"]
        self.spotlight: Optional[Spotlight] = get_optional(data, "spotlight", Spotlight)
        self.beatmapsets: Optional[List[Beatmapset]] = get_optional_list(data, "beatmapsets", Beatmapset)

    def __repr__(self):
        return prettify(self, "ranking")


class Spotlight:
    """
    The details of a spotlight.

    **Attributes**

    end_date: :class:`datetime.datetime`
        In DateTime format. The end date of the spotlight.

    id: :class:`int`
        The ID of this spotlight.

    mode_specific: :class:`bool`
        If the spotlight has different modes specific to each game mode.

    participant_count: Optional[:class:`int`]
        The number of users participating in this spotlight. This is only shown when viewing a single spotlight.

    name: :class:`str`
        The name of the spotlight.

    start_date: :class:`datetime.datetime`
        In DatTime format. The starting date of the spotlight.

    type: :class:`str`
        The type of spotlight.
    """

    __slots__ = (
        "end_date",
        "id",
        "mode_specific",
        "name",
        "start_date",
        "type",
        "participant_count",
    )

    def __init__(self, data):
        self.end_date: datetime = parser.parse(data["end_date"])
        self.id: int = data["id"]
        self.mode_specific: bool = data["mode_specific"]
        self.participant_count: Optional[int] = data.get("participant_count")
        self.name: str = data["name"]
        self.start_date: datetime = parser.parse(data["start_date"])
        self.type: str = data["type"]

    def __repr__(self):
        return prettify(self, "name")


class Spotlights:
    """
    **Attributes**

    spotlights: List[:class:`Spotlight`]
    """

    __slots__ = ("spotlights",)

    def __init__(self, data):
        self.spotlights: List[Spotlight] = list(map(Spotlight, data["spotlights"]))

    def __repr__(self):
        return prettify(self, "spotlights")



from dateutil import parser
from typing import Optional, List, TYPE_CHECKING

from ..util import prettify, get_optional
from ..enums import ForumTopicType


if TYPE_CHECKING:
    from datetime import datetime


class ForumPost:
    """
    **Attributes**

    created_at: :class:`datetime.datetime`

    deleted_at: Optional[:class:`datetime.datetime`]

    edited_at: Optional[:class:`datetime.datetime`]

    edited_by_id: Optional[:class:`int`]

    forum_id: :class:`int`

    id: :class:`int`

    topic_id: :class:`int`

    user_id: :class:`int`

    body: Optional[:class:`TextFormat`]
    """

    __slots__ = (
        "created_at",
        "deleted_at",
        "edited_at",
        "edited_by_id",
        "forum_id",
        "id",
        "topic_id",
        "user_id",
        "body",
    )

    def __init__(self, data):
        self.created_at: datetime = parser.parse(data["created_at"])
        self.deleted_at: Optional[datetime] = get_optional(data, "deleted_at", parser.parse)
        self.edited_at: Optional[datetime] = get_optional(data, "edited_at", parser.parse)
        self.edited_by_id: Optional[int] = data.get("edited_by_id")
        self.forum_id: int = data["forum_id"]
        self.id: int = data["id"]
        self.topic_id: int = data["topic_id"]
        self.user_id: int = data["user_id"]
        self.body: Optional[TextFormat] = get_optional(data, "body", TextFormat)

    def __repr__(self):
        return prettify(self, "user_id", "topic_id")


class TextFormat:
    """
    A page that has html and raw data

    **Attributes**

    html: :class:`str`

    raw: :class:`str`
    """

    __slots__ = ("html", "raw")

    def __init__(self, data):
        self.html: str = data["html"]
        for attr in ("raw", "bbcode", "markdown"):
            if attr in data:
                self.raw: str = data[attr]
                break
        if self.raw is None:
            raise KeyError("data did not contain a raw value")

    def __repr__(self):
        return prettify(self, "raw")


class ForumTopic:
    """
    **Attributes**

    created_at: :class:`datetime.datetime`

    deleted_at: Optional[:class:`datetime.datetime`]

    first_post_id: :class:`int`

    forum_id: :class:`int`

    id: :class:`int`

    is_locked: :class:`bool`

    last_post_id: :class:`int`

    poll: Optional[:class:`Poll`]

    post_count: :class:`int`

    title: :class:`str`

    type: :class:`ForumTopicType`

    updated_at: :class:`datetime.datetime`

    user_id: :class:`int`
    """

    __slots__ = (
        "created_at",
        "deleted_at",
        "first_post_id",
        "forum_id",
        "id",
        "is_locked",
        "last_post_id",
        "poll",
        "post_count",
        "title",
        "type",
        "updated_at",
        "user_id",
    )

    def __init__(self, data):
        self.created_at: datetime = parser.parse(data["created_at"])
        self.deleted_at: Optional[datetime] = get_optional(data, "deleted_at", parser.parse)
        self.first_post_id: int = data["first_post_id"]
        self.forum_id: int = data["forum_id"]
        self.id: int = data["id"]
        self.is_locked: bool = data["is_locked"]
        self.last_post_id: int = data["last_post_id"]
        self.poll: Optional[Poll] = get_optional(data, "poll", Poll)
        self.post_count: int = data["post_count"]
        self.title: str = data["title"]
        self.type: ForumTopicType = ForumTopicType(data["type"])
        self.updated_at: datetime = parser.parse(data["updated_at"])
        self.user_id: int = data["user_id"]

    def __repr__(self):
        return prettify(self, "user_id", "title")


class Poll:
    """
    **Attributes**

    allow_vote_change: :class:`bool`

    ended_at: Optional[:class:`datetime.datetime`]

    hide_incomplete_results: :class:`bool`

    last_vote_at: Optional[:class:`datetime.datetime`]

    max_votes: :class:`int`

    options: List[:class:`PollOption`]

    started_at: :class:`datetime.datetime`

    title: :class:`TextFormat`

    total_vote_count: :class:`int`
    """

    __slots__ = (
        "allow_vote_change",
        "ended_at",
        "hide_incomplete_results",
        "last_vote_at",
        "max_votes",
        "options",
        "started_at",
        "title",
        "total_vote_count",
    )

    def __init__(self, data):
        self.allow_vote_change: bool = data["allow_vote_change"]
        self.ended_at: Optional[datetime] = get_optional(data, "ended_at", parser.parse)
        self.hide_incomplete_results: bool = data["hide_incomplete_results"]
        self.last_vote_at: Optional[datetime] = get_optional(data, "last_vote_at", parser.parse)
        self.max_votes: int = data["max_votes"]
        self.options: List[PollOption] = list(map(PollOption, data["options"]))
        self.started_at: datetime = parser.parse(data["started_at"])
        self.title: TextFormat = TextFormat(data["title"])
        self.total_vote_count: int = data["total_vote_count"]

    def __repr__(self):
        return prettify(self, "title", "options")


class PollOption:
    """
    **Attributes**

    id: :class:`int`

    text: :class:`TextFormat`

    vote_count: Optional[:class:`int`]
        Not present if the poll is incomplete and results are hidden.
    """

    __slots__ = ("id", "text", "vote_count")

    def __init__(self, data):
        self.id: int = data["id"]
        self.text: TextFormat = TextFormat(data["text"])
        self.vote_count: Optional[int] = data["vote_count"]

    def __repr__(self):
        return prettify(self, "text", "vote_count")
